/*
 * Copyright 2016, Fabrice DESCLAUX <serpilliere@droids-corp.org>
 * Copyright 2016, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <string.h>

#include "ucg_reent_intr.h"

struct saved_stack {
	uint32_t r0;
	uint32_t r1;
	uint32_t r2;
	uint32_t r3;
	uint32_t r12;
	uint32_t lr;
	uint32_t pc;
	uint32_t psr;
};

struct wrapper_stack {
	uint32_t psr;
	uint32_t lr;
	uint32_t r0;
	uint32_t r1;
	uint32_t r2;
	uint32_t r3;
	uint32_t r12;
	uint32_t pc;
};

__attribute__((naked)) uint32_t
ret_wrapper_stkalign_float(void)
{
	__asm__ volatile (
		/* restore float */
		"ADD    SP, SP, 0x50       \n"
		/* restore PSR, LR */
		"POP   {R0, LR}        \n"
		/* restore PSR */
		//"MSR   PSR, R0       \n"
		"msr APSR_nzcvq, R0 \n"
		/* restore PSR, LR */
		"POP   {R0-R3, R12, PC}        \n"
	);
}


__attribute__((naked)) uint32_t
ret_wrapper_stknoalign_float(void)
{
	__asm__ volatile (
		/* restore float */
		"ADD    SP, SP, 0x50       \n"
		/* remove padding */
		"ADD    SP, SP, 0x4       \n"
		/* restore PSR, LR */
		"POP   {R0, LR}        \n"
		/* restore PSR */
		//"MSR   PSR, R0       \n"
		"msr APSR_nzcvq, R0 \n"
		/* restore PSR, LR */
		"POP   {R0-R3, R12, PC}        \n"
	);
}


__attribute__((naked)) uint32_t
ret_wrapper_stkalign_nofloat(void)
{
	__asm__ volatile (
		/* restore float */
		"ADD    SP, SP, 0x8       \n"
		/* restore PSR, LR */
		"POP   {R0, LR}        \n"
		/* restore PSR */
		//"MSR   PSR, R0       \n"
		"msr APSR_nzcvq, R0 \n"
		/* restore PSR, LR */
		"POP   {R0-R3, R12, PC}        \n"
	);
}

__attribute__((naked)) uint32_t
ret_wrapper_stknoalign_nofloat(void)
{
	__asm__ volatile (
		/* restore float */
		"ADD    SP, SP, 0x8       \n"
		/* remove padding */
		"ADD    SP, SP, 0x4       \n"
		/* restore PSR, LR */
		"POP   {R0, LR}        \n"
		/* restore PSR */
		//"MSR   PSR, R0       \n"
		"msr APSR_nzcvq, R0 \n"
		/* restore PSR, LR */
		"POP   {R0-R3, R12, PC}        \n"
	);
}

void reent_intr(void);

void ucg_reent_intr(uint32_t *context, void *fct)
//void ucg_reent_intr(uint32_t *context)
{
	struct saved_stack *stk_new = (void *)context;
	struct saved_stack *stk_old = (void *)(&context[10]);
	uint32_t exe_return;
	int stk_padding;
	struct wrapper_stack wstack;

	stk_padding = (stk_old->psr & 0x200)?1:0;

	/* copy saved ctxt */
	wstack.r0 = stk_old->r0;
	wstack.r1 = stk_old->r1;
	wstack.r2 = stk_old->r2;
	wstack.r3 = stk_old->r3;
	wstack.r12 = stk_old->r12;
	wstack.lr = stk_old->lr;
	wstack.psr = stk_old->psr & ~0x200;
	wstack.pc = stk_old->pc | 1;

	exe_return = (uint32_t)context[-4];


	stk_new->pc = (uint32_t)fct;
//	stk_new->pc = (uint32_t)reent_intr;
	/* set arguments */
	/* stk_new->r0 = 0; */
	/* stk_new->r1 = 1; */
	/* stk_new->r2 = 2; */
	/* stk_new->r3 = 3; */
	stk_new->r12 = wstack.r12;
	stk_new->psr = 0x21000000;

	if (exe_return & 0x10) {
		/*  No Float stacked */
		if (stk_padding) {
			stk_new->lr = (uint32_t)ret_wrapper_stknoalign_nofloat;
			memcpy(&context[10+1], &wstack, sizeof(wstack));
		} else {
			memcpy(&context[10], &wstack, sizeof(wstack));
			stk_new->lr = (uint32_t)ret_wrapper_stkalign_nofloat;
		}
	} else {
		/* Float stacked */
		//STM_EVAL_LEDOn(LED7);
		memmove(&context[10], &context[18], 18*4);
		if (stk_padding) {
			//STM_EVAL_LEDOn(LED10);
			memcpy(&context[10+18+1], &wstack, sizeof(wstack));
			stk_new->lr = (uint32_t)ret_wrapper_stknoalign_float;
		} else {
			//STM_EVAL_LEDOn(LED9);
			memcpy(&context[10+18], &wstack, sizeof(wstack));
			stk_new->lr = (uint32_t)ret_wrapper_stkalign_float;
		}

	}

	/* force return no fpu */
	context[-4] |= 0x10;
	/* create exe_return for real state return */
	context[8] = 0x1337beef;
	context[8+1] = exe_return;
}
